##############################################################################
#
# Copyright (c) 2010, 2011, 2012 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from __future__ import annotations

import configparser
import errno
import ipaddress
import logging
import netifaces
import os
import pwd
import grp
import subprocess
import sys
import time

from collections import defaultdict
from netifaces import AF_INET, AF_INET6


def do_format(conf):
  # starting logs
  conf.logStartInfo()
  # wrap system calls for tracing and dry-run
  with WrappedSystem(conf):
    # load configuration
    computer = Computer(conf)
    # sanity checks
    computer.checkConf()
    # format
    computer.format()
    # check formatting
    computer.checkFormat()
    # collect some environmental data
    computer.update()
    # send to master
    computer.send()


class UsageError(Exception):
  def __init__(self, message):
    self.message = message
    Exception.__init__(self, message)


class Parameters(object):
  master_url: str
  computer_id: str
  instance_root: str
  software_root: str
  partition_amount: int


class Options(object):
  input_definition_file: str = None
  output_definition_file: str = None
  computer_cache_file: str = None

  key_file: str = None
  cert_file: str = None

  log_file: str = None
  dry_run: bool = False

  software_user = 'slapsoft'
  partition_base_name = 'slappart'
  user_base_name = 'slapuser'

  interface_name: str = None
  ipv6_interface: str = None
  ipv4_local_network: str = None
  ipv6_range: bool = True

  create_tap: bool = True
  tap_base_name: str = 'slaptap'
  tap_ipv6: bool = True
  tap_ipv4_prefix: int = 32
  tap_gateway_interface: str = None

  create_tun: bool = False
  tun_base_name: str = 'slaptun'
  tun_ipv6: bool = True
  tun_ipv4_network: str = '172.16.0.0/12'


class FormatConfig(Parameters, Options):
  DEPRECATED = [
    'bridge_name',
    'no_bridge_',
    'master_ca_file',
    'alter_network',
    'alter_user',
    'computer_json',
    'computer_xml',
  ]
  CHECK_FILES = ['key_file', 'cert_file', 'input_definition_file']
  NORMALIZE_PATHS = ['instance_root', 'software_root']

  logger : logging.Logger

  def __init__(self):
    self.logger = logging.getLogger("slapos.format")

  def abort(self, fmt, *args, exc=ValueError):
    message = fmt % tuple(args)
    self.logger.error(message)
    raise exc(message)

  def wrong(self, fmt, *args):
    return self.abort(fmt, *args, exc=UsageError)

  def dryError(self, fmt, *args):
    self.logger.warning('[dry-error] ' + fmt, *args)

  def warn(self, fmt, *args):
    self.logger.warning(fmt, *args)

  def info(self, fmt, *args):
    self.logger.info(fmt, *args)

  def debug(self, fmt, *args):
    self.logger.debug(fmt, *args)

  def parse(self, name, value, t):
    if not isinstance(value, str):
      if type(value).__name__ != t and value is not None:
        self.wrong("Option %s takes type %s, not %r", name, t, value)
      return value
    if t == 'int':
      try:
        return int(value)
      except ValueError:
        self.wrong("Option %s takes type %s, not %r", name, t, value)
    if t == 'bool':
      try:
        return bool(('false', 'true').index(value.lower()))
      except IndexError:
        self.wrong("Option %r must be 'true' or 'false', not %r", name, value)
    return value

  def get(self, option):
    try:
      return getattr(self, option)
    except AttributeError:
      self.wrong("Parameter %r is not defined", option)

  def mergeConfig(self, args, configp):
    # args (from command line) override configp (from cfg) options
    for section in ('slapformat', 'slapos'):
      self.__dict__.update(configp.items(section))
    self.__dict__.update(args.__dict__)

  def setConfig(self):
    # Deprecated options
    for option in self.DEPRECATED:
      if option in self.__dict__:
        if option == 'computer_xml':
          self.wrong(
            "Option %r is no longer supported\n"
            "Use --output_definition_file to migrate existing computer_xml"
            ", then use the generated file as input_definition_file",
            option
          )
        else:
          self.wrong("Option %r is no longer supported" % option)
    # Mandatory parameters
    for option, t in Parameters.__annotations__.items():
      setattr(self, option, self.parse(option, self.get(option), t))
    # Optional inputs
    for option, t in Options.__annotations__.items():
      setattr(self, option, self.parse(option, getattr(self, option), t))
    # Existence checks for files
    for option in self.CHECK_FILES:
      path = getattr(self, option)
      if path:
        if not os.path.exists(path):
          self.wrong("File %r does not exist or is not readable", path)
        setattr(self, option, os.path.abspath(path))
    # Paths normalization
    for option in self.NORMALIZE_PATHS:
      path = getattr(self, option)
      if path:
        setattr(self, option, os.path.abspath(path))
    # XXX Check command line tools

  def logStartInfo(self):
    if self.dry_run:
      self.logger.info("Starting run in dry-run mode")
    else:
      self.logger.info("Starting run")
    # XXX more logs


class Definition(defaultdict):
  def __init__(self, definition_file):
    super().__init__(lambda: defaultdict(type(None)))
    if definition_file:
      configp = configparser.ConfigParser(interpolation=None)
      configp.read(definition_file)
      for s in configp.sections():
        self[s] = defaultdict(type(None), configp.items(s))


class Computer(object):
  conf : FormatConfig
  reference : str
  interface : Interface
  partitions : list[Partition]
  address : ipaddress.IPv4Interface | ipaddress.IPv6Interface
  user : User

  def __init__(self, conf):
    definition = Definition(conf.input_definition_file)
    section = definition['computer']
    # Basic configuration
    self.conf = conf
    self.reference = conf.computer_id
    # Interface
    self.interface = Interface(conf)
    # Address
    addr = section['address']
    address = ipaddress.ip_interface(addr) if addr else None
    self.address = address or self.interface.getComputerIPv6Addr()
    # User
    username = section['software_user'] or conf.software_user
    path = conf.software_root
    self.user = User(username, path)
    # Partitions
    amount = conf.partition_amount
    self.partitions = [Partition(i, self, definition) for i in range(amount)]

  def checkConf(self):
    conf = self.conf
    interface = self.interface
    partitions =  self.partitions
    conf.info("Checking configuration")
    # Big enough IPv6 network
    if any(p.ipv6_range for p in partitions):
      interface.checkPartitionIPv6Ranges()
    if any(p.tap and p.tap.ipv6_gateway for p in partitions):
      interface.checkTapIPv6Ranges()
    # Warn about IP address/ranges overlaps
    self.checkAddressOverlaps()

  def checkAddressOverlaps(self):
    def single(ip):
      return ipaddress.ip_network((ip.ip, ip.max_prefixlen))
    ipv4 = []
    ipv6 = []
    ipv4_maybe_shared = {}
    ipv6_maybe_shared = {}
    if self.address.version == 4:
      ipv4_maybe_shared[single(self.address)] = self.reference
    else:
      ipv6_maybe_shared[single(self.address)] = self.reference
    for p in self.partitions:
      ipv4.extend((single(a), p.reference + ' ip') for a in p.ipv4_list)
      ipv6.extend((single(a), p.reference + ' ip') for a in p.ipv6_list)
      if p.ipv6_range:
        ipv6.append((p.ipv6_range, p.reference + ' ipv6 range'))
      if p.tap:
        dev = p.tap.name
        if p.tap.ipv4_address:
          ipv4.append((p.tap.ipv4_address.network, dev))
          ipv4_maybe_shared[single(p.tap.ipv4_gateway)] = dev + ' gateway'
        if p.tap.ipv6_gateway:
          ipv6.append((p.tap.ipv6_gateway.network, dev))
          ipv6_address = p.tap.ipv6_address
          if ipv6_address.ip not in p.tap.ipv6_gateway.network:
            ipv6_maybe_shared[single(ipv6_address.network)] = dev + ' address'
    ipv4.extend(ipv4_maybe_shared.items())
    ipv6.extend(ipv6_maybe_shared.items())
    for range_list in (ipv4, ipv6):
      range_list.sort()
      it = iter(range_list)
      r1, reason1 = next(it)
      for r2, reason2 in it:
        if r1.overlaps(r2):
          self.conf.warn(
            "%s (%s) and %s (%s) overlap",
            r1, reason1, r2, reason2
          )
        r1, reason1 = r2, reason2

  def format(self):
    self.conf.info("Formatting (%d partitions)", len(self.partitions))
    # Software root path & user
    ensureDir(self.conf.software_root, 0o755)
    self.user.ensureUser()
    uid, gid = self.user.getIds()
    os.chown(self.conf.software_root, uid, gid)
    # Non local bind for IPv6 ranges
    if self.conf.ipv6_range or any(p.ipv6_range for p in self.partitions):
      self.interface.enableIPv6NonLocalBind()
    # Computer address
    self.interface.addAddress(self.address, self.reference)
    # Instance root path
    ensureDir(self.conf.instance_root, 0o755)
    # Format each partitions
    for p in self.partitions:
      p.format(self.interface)

  def checkFormat(self):
    self.conf.info("Checking formatting")
    time.sleep(2)
    for p in self.partitions:
      p.checkFormat(self.interface)

  def update(self):
    pass

  def send(self):
    pass


class Partition(object):
  reference : str
  index: int
  path : str
  user : User
  ipv4_list: list[ipaddress.IPv4Interface]
  ipv6_list: list[ipaddress.IPv6Interface]
  ipv6_range: ipaddress.IPv6Network = None
  tap : Tap = None
  tun : Tun = None

  def __init__(self, index, computer, definition=None):
    i = str(index)
    conf = computer.conf
    interface = computer.interface
    section = definition['partition_' + i]
    options = defaultdict(type(None), section, **definition['default'])
    # Reference, path & user
    self.reference = options['pathname'] or conf.partition_base_name + i
    self.path = os.path.join(conf.instance_root, self.reference)
    self.user = User(options['user'] or conf.user_base_name + i, self.path)
    # IPv4 & IPv6 addresses
    ipv4_list = ipv6_list = None
    addresses = options['address']
    if addresses:
      address_list = [ipaddress.ip_interface(a) for a in addresses.split(' ')]
      ipv4_list = [ip for ip in address_list if ip.version == 4]
      ipv6_list = [ip for ip in address_list if ip.version == 6]
    self.ipv4_list = ipv4_list or [interface.getPartitionIPv4Addr(index)]
    self.ipv6_list = ipv6_list or [interface.getPartitionIPv6Addr(index)]
    # IPv6 range
    ipv6_range = options['ipv6_range']
    if ipv6_range:
      self.ipv6_range = ipaddress.IPv6Network(ipv6_range, strict=False)
    elif conf.ipv6_range:
      self.ipv6_range = interface.getIPv6Range(index)
    # Tap
    tap = options['network_interface']
    if tap or conf.create_tap:
      tap = tap or conf.tap_base_name + i
      self.tap = Tap(tap, index, computer, definition)
    # Tun
    # XXX

  def format(self, interface):
    self.user.ensureUser()
    self.ensurePath()
    for ip in self.ipv4_list:
      interface.addAddress(ip, self.reference)
    for ip in self.ipv6_list:
      interface.addAddress(ip, self.reference)
    if self.tap:
      self.tap.format(interface, self.user)

  def checkFormat(self, interface):
    for ip in self.ipv4_list:
      interface.checkAddress(ip, self.reference)
    for ip in self.ipv6_list:
      interface.checkAddress(ip, self.reference)

  def ensurePath(self):
    if not os.path.exists(self.path):
      os.mkdir(self.path)
    uid, gid = self.user.getIds()
    os.chown(self.path, uid, gid)
    os.chmod(self.path, 0o750)


class User(object):
  name : str
  path : str
  groups : list[str]

  SHELL = '/bin/sh'

  def __init__(self, name, path, groups=None):
    self.name = name
    self.path = path
    self.groups = groups

  def ensureUser(self):
    groupname = 'grp_' + self.name if sys.platform == 'cygwin' else self.name
    if not self.groupExists(groupname):
      call(['groupadd', groupname])
    parameters = ['-d', self.path, '-g', self.name, '-s', self.SHELL]
    if self.groups:
      parameters.extend(['-G', ','.join(self.groups), '-a'])
    parameters.append(self.name)
    if self.userExists(self.name):
      # if the user is already created and used we should not fail
      call(['usermod'] + parameters, throw=False)
    else:
      parameters.append('-r')
      call(['useradd'] + parameters)
    # lock the password of user
    call(['passwd', '-l', self.name])

  def getIds(self):
    pw = pwd.getpwnam(self.name)
    return pw.pw_uid, pw.pw_gid

  @classmethod
  def groupExists(cls, name):
    try:
      grp.getgrnam(name)
      return True
    except KeyError:
      return False

  @classmethod
  def userExists(cls, name):
    try:
      pwd.getpwnam(name)
      return True
    except KeyError:
      return False


class Tap(object):
  name : str
  ipv4_gateway : ipaddress.IPv4Interface = None
  ipv4_address : ipaddress.IPv4Interface = None
  ipv6_gateway : ipaddress.IPv6Interface = None
  ipv6_address : ipaddress.IPv6Interface = None

  def __init__(self, name, index, computer, definition):
    conf = computer.conf
    interface = computer.interface
    options = definition[name]
    self.name = name
    # IPv4 (gateway=host and address=application)
    ipv4_gateway = options['ipv4_gateway']
    ipv4_address = options['ipv4_address']
    if bool(ipv4_gateway) != bool(ipv4_address):
      conf.wrong(
        "Options %r and %r for tap %s must either both be provided or neither",
        'ipv4_gateway', 'ipv4_address', name
      )
    if ipv4_address:
      self.ipv4_gateway = ipaddress.IPv4Interface(ipv4_gateway)
      self.ipv4_address = ipaddress.IPv4Interface(ipv4_address)
    elif interface.tap_ipv4_network:
      self.ipv4_gateway = interface.getTapIPv4Gateway(index)
      self.ipv4_address = interface.getTapIPv4Range(index)
    # IPv6 (gateway=application and address=host)
    ipv6_gateway = options['ipv6_gateway']
    ipv6_address = options['ipv6_address']
    if bool(ipv6_gateway) != bool(ipv6_address):
      conf.wrong(
        "Options %r and %r for tap %s must either both be provided or neither",
        'ipv6_address', 'ipv6_gateway', name
      )
    if ipv6_gateway:
      self.ipv6_gateway = ipaddress.IPv6Interface(ipv6_gateway)
      self.ipv6_address = ipaddress.IPv6Interface(ipv6_address)
    elif conf.tap_ipv6:
      ipv6_network = interface.getIPv6Range(index, 'tap')
      self.ipv6_gateway = addressFromNetwork(ipv6_network, 1)
      self.ipv6_address = addressFromNetwork(ipv6_network, -1)

  def format(self, interface, user):
    conf = interface.conf
    # Create/ensure tap interface
    self.ensureTap(user, conf)
    # Create/ensure IPv4 routes
    self.ensureIPv4Routes(conf)
    # Add host IPV6 address to interface
    if self.ipv6_address:
      interface.addIPv6(self.ipv6_address)
    # Create/ensure IPv6 routes
    self.ensureIPv6Routes(conf)

  def ensureTap(self, user, conf):
    tap = self.name
    user = user.name
    check_file = '/sys/devices/virtual/net/%s/owner' % tap
    if os.path.exists(check_file):
      with open(check_file) as f:
        owner_id = f.read().strip()
      try:
        owner_id = int(owner_id)
      except ValueError:
        owner_id = pwd.getpwnam(owner_id).pw_uid
      user_id = pwd.getpwnam(user).pw_uid
      if owner_id != user_id:
        conf.abort(
          "Wrong owner of tap %s, expected %s (%s) got %s",
          tap, user_id, user, owner_id
        )
    else:
      call(['ip', 'tuntap', 'add', 'dev', tap, 'mode', 'tap', 'user', user])
      call(['ip', 'link', 'set', tap, 'up'])

  def ensureIPv4Routes(self, conf):
    tap = self.name
    ipv4 = self.ipv4_address
    if ipv4:
      cidr = str(ipv4)
      code, out = call(['ip', 'route', 'show', cidr], throw=False)
      if code != 0 or cidr not in out or tap not in out:
        call(['ip', 'route', 'add', cidr, 'dev', tap])
      if ipv4.network.prefixlen < 32:
        network = str(ipv4.network)
        code, out = call(['ip', 'route', 'show', network], throw=False)
        if code != 0 or 'dev ' + tap not in out:
          call(['ip', 'route', 'add', network, 'dev', tap, 'via', cidr])
        elif 'via ' + cidr not in out:
          conf.abort(
            "Route to network %s (%s) exists but is not via %s",
            network, tap, cidr
          )

  def ensureIPv6Routes(self, conf):
    tap = self.name
    ipv6 = self.ipv6_gateway
    if ipv6:
      cidr = str(ipv6)
      code, out = call(['ip', '-6', 'route', 'show', cidr], throw=False)
      if code != 0 or cidr not in out or tap not in out:
        call(['ip', '-6', 'route', 'add', cidr, 'dev', tap])
      network = str(ipv6.network)
      code, out = call(['ip', '-6', 'route', 'show', network], throw=False)
      if code != 0 or 'dev ' + tap not in out:
        call(['ip', '-6', 'route', 'add', network, 'dev', tap, 'via', cidr])
      elif 'via ' + cidr not in out:
        conf.warn(
          "Route to network %s (%s) exists but is not via %s, fixing it",
          network, tap, cidr
        )
        call(['ip', '-6', 'route', 'del', network, 'dev', tap])
        call(['ip', '-6', 'route', 'add', network, 'dev', tap, 'via', cidr])


class Tun(object):
  name : str
  ipv4_address : ipaddress.IPv4Interface = None
  ipv6_address : ipaddress.IPv6Interface = None

  def __init__(self, name, index, computer, definition):
    conf = computer.conf
    interface = computer.interface
    options = definition[name]
    self.name = name
    # IPv4
    ipv4_address = options['ipv4_address']
    if ipv4_address:
      self.ipv4_address = ipaddress.IPv4Interface(ipv4_address)
    else:
      self.ipv4_address = interface.getTunIPv4Range(index)
    # IPv6
    ipv6_address = options['ipv6_address']
    if ipv6_address:
      self.ipv6_address = ipaddress.IPv6Interface(ipv6_address)
    elif conf.tap_ipv6:
      ipv6_network = interface.getIPv6Range(index, 'tun')
      self.ipv6_address = addressFromNetwork(ipv6_network, -1)

  def format(self, interface, user):
    conf = interface.conf
    # Create/ensure tun interface
    self.ensureTun(user, conf)
    # Create/ensure IPv4 routes
    self.ensureIPv4Routes(conf)
    # Create/ensure IPv6 routes
    self.ensureIPv6Routes(conf)

  def ensureTun(self, user, conf):
    tun = self.name
    user = user.name
    check_file = '/sys/devices/virtual/net/%s/owner' % tun
    if os.path.exists(check_file):
      with open(check_file) as f:
        owner_id = f.read().strip()
      try:
        owner_id = int(owner_id)
      except ValueError:
        owner_id = pwd.getpwnam(owner_id).pw_uid
      user_id = pwd.getpwnam(user).pw_uid
      if owner_id != user_id:
        conf.abort(
          "Wrong owner of tun %s, expected %s (%s) got %s",
          tun, user_id, user, owner_id
        )
    else:
      call(['ip', 'tuntap', 'add', 'dev', tun, 'mode', 'tun', 'user', user])
      call(['ip', 'link', 'set', tun, 'up'])

  def ensureIPv4Routes(self, conf):
    tun = self.name
    ipv4 = self.ipv4_address
    if ipv4:
      cidr = str(ipv4)
      code, out = call(['ip', 'route', 'show', cidr], throw=False)
      if code != 0 or cidr not in out or tun not in out:
        call(['ip', 'route', 'add', cidr, 'dev', tun])
      if ipv4.network.prefixlen < 32:
        network = str(ipv4.network)
        code, out = call(['ip', 'route', 'show', network], throw=False)
        if code != 0 or 'dev ' + tun not in out:
          call(['ip', 'route', 'add', network, 'dev', tun, 'via', cidr])
        elif 'via ' + cidr not in out:
          conf.abort(
            "Route to network %s (%s) exists but is not via %s",
            network, tap, cidr
          )

  def ensureIPv6Routes(self, conf):
    tap = self.name
    ipv6 = self.ipv6_gateway
    if ipv6:
      cidr = str(ipv6)
      code, out = call(['ip', '-6', 'route', 'show', cidr], throw=False)
      if code != 0 or cidr not in out or tap not in out:
        call(['ip', '-6', 'route', 'add', cidr, 'dev', tap])
      network = str(ipv6.network)
      code, out = call(['ip', '-6', 'route', 'show', network], throw=False)
      if code != 0 or 'dev ' + tap not in out:
        call(['ip', '-6', 'route', 'add', network, 'dev', tap, 'via', cidr])
      elif 'via ' + cidr not in out:
        conf.warn(
          "Route to network %s (%s) exists but is not via %s, fixing it",
          network, tap, cidr
        )
        call(['ip', '-6', 'route', 'del', network, 'dev', tap])
        call(['ip', '-6', 'route', 'add', network, 'dev', tap, 'via', cidr])


class Interface(object):
  ipv4_interface : str
  ipv6_interface : str
  ipv4_network : ipaddress.IPv4Network
  ipv6_network : ipaddress.IPv6Network
  tap_ipv4_network : ipaddress.IPv4Network
  conf : FormatConfig

  def __init__(self, conf):
    self.conf = conf
    self.ipv4_interface = conf.interface_name
    self.ipv6_interface = conf.ipv6_interface or conf.interface_name
    self.ipv4_network = self.getIPv4Network(conf.ipv4_local_network)
    self.ipv6_network = self.getIPv6Network()
    self.tap_ipv4_network = self.getTapIPv4Network(conf)
    # Check IPv4 networks overlap
    tap_network = self.tap_ipv4_network
    ipv4_network = self.ipv4_network
    if ipv4_network and tap_network and ipv4_network.overlaps(tap_network):
      self.conf.abort(
        "IPv4 network %s and tap IPv4 network %s overlap",
        ipv4_network, tap_network
      )

  def checkPartitionIPv6Ranges(self):
    if self.ipv6_network.prefixlen > 128 - 16:
      self.conf.abort("IPv6 network %s is too small for IPv6 ranges", network)

  def checkTapIPv6Ranges(self):
    if self.ipv6_network.prefixlen > 128 - 16 - 1:
      self.conf.abort("IPv6 network %s is too small for tap ranges", network)

  def getIPv4Network(self, cidr):
    if cidr:
      # XXX allow ipv4_local_network to be None ?
      return ipaddress.IPv4Network(cidr, strict=False)

  def getPartitionIPv4Addr(self, index):
    return addressFromNetwork(self.ipv4_network, index + 2)

  def getTapIPv4Network(self, conf):
    tap_gateway_interface = conf.tap_gateway_interface
    if conf.create_tap and tap_gateway_interface:
      try:
        addresses = netifaces.ifaddresses(tap_gateway_interface)[AF_INET]
      except ValueError:
        self.conf.abort(
          "Interface %s (intended as tap gateway interface) does not exist",
          tap_gateway_interface
        )
      except KeyError:
        self.conf.abort(
          "%s must have at least one IPv4 address assigned",
          tap_gateway_interface
        )
      result = None
      for a in addresses:
        address = a['addr'].split('%')[0]
        netmask = a['netmask'].split('/')[-1]
        ip = ipaddress.IPv4Interface('%s/%s' % (address, netmask))
        network = ip.network
        if not result or network.prefixlen < result.prefixlen:
          result = network
      return result

  def getTapIPv4Gateway(self, _index):
    network = self.tap_ipv4_network
    if network:
      return addressFromNetwork(network, 1)

  def getTapIPv4Range(self, index):
    network = self.tap_ipv4_network
    if network:
      prefixlen = conf.tap_ipv4_prefix
      i = index + 2 if prefixlen == 32 else index + 1
      k = 1 if prefixlen < 31 else 0
      addr = network[(i << (32 - prefixlen)) + k]
      return ipaddress.IPv4Interface((addr, prefixlen))

  def getIPv6Network(self):
    try:
      addresses = netifaces.ifaddresses(self.ipv6_interface)[AF_INET6]
    except KeyError:
      self.conf.abort(
        "%s must have at least one IPv6 address assigned",
        self.ipv6_interface
      )
    result = None
    for a in addresses:
      address = a['addr'].split('%')[0]
      netmask = a['netmask'].split('/')[-1]
      ip = ipaddress.IPv6Interface('%s/%s' % (address, netmask))
      network = ip.network
      if network.is_global:
        if not result or network.prefixlen < result.prefixlen:
          result = network
    return result

  def getComputerIPv6Addr(self):
    return addressFromNetwork(self.ipv6_network, 1)

  def getPartitionIPv6Addr(self, index):
    return addressFromNetwork(self.ipv6_network, index + 2)

  def getIPv6Range(self, index, mode='partition'):
    k = ('partition', 'tap', 'tun').index(mode) + 1
    network = self.ipv6_network
    prefixlen = network.prefixlen + 16
    bits = 128 - network.prefixlen
    addr = network[(k << (bits - 2)) + (index << (128 - prefixlen))]
    return ipaddress.IPv6Network((addr, prefixlen))

  def enableIPv6NonLocalBind(self):
    call(['sysctl', 'net.ipv6.ip_nonlocal_bind=1'])
    network = str(self.ipv6_network)
    _, result = call(['ip', '-6', 'route', 'show', 'table', 'local', network])
    if not 'dev lo' in result:
      call(['ip', '-6', 'route', 'add', 'local', network, 'dev', 'lo'])

  def addAddress(self, ip, reason, interface=None):
    interface = interface or getattr(self, 'ipv%d_interface' % ip.version)
    af = {4: AF_INET, 6: AF_INET6}[ip.version]
    address = str(ip.ip)
    netmask = str(ip.network.netmask)
    for iface in netifaces.interfaces():
      if iface != interface:
        ifaddresses = netifaces.ifaddresses(iface).get(af, ())
        if address in (q['addr'].split('%')[0] for q in ifaddresses):
          self.conf.abort(
            "Cannot add address %s (%s) to %s because it already exists on %s",
            ip, reason, interface, iface
          )
    ifaddresses = netifaces.ifaddresses(interface).get(af, ())
    for q in ifaddresses:
      if address == q['addr'].split('%')[0]:
        if netmask not in q['netmask']:
           self.conf.abort(
            "Address %s (%s) is already on %s but with another netmask %s",
            ip, reason, interface, q['netmask']
          )
        break
    else:
      call(['ip', 'addr', 'add', str(ip), 'dev', interface])

  def checkAddress(self, ip, reason, interface=None):
    interface = interface or getattr(self, 'ipv%d_interface' % ip.version)
    flag = str(-ip.version)
    _, result = call(['ip', flag, 'addr', 'show', interface])
    cidr = str(ip)
    for line in result.splitlines():
      if cidr in line:
        for state in ('tentative', 'dadfailed'):
          if state in line:
            call(['ip', 'addr', 'del', cidr, 'dev', interface])
            self.conf.abort(
              "Address %s (%s) on %s was in state %s so was removed: %r",
              cidr, reason, interface, state, line
            )
        return
    self.conf.abort(
      "Address %s (%s) is unexpectedly not present on %s",
      cidr, reason, interface
    )


# Utilities

def call(args, throw=True):
  popen = subprocess.Popen(
    args,
    stdout=subprocess.PIPE,
    stderr=subprocess.STDOUT,
    universal_newlines=True,
  )
  result = popen.communicate()[0]
  if throw and popen.returncode != 0:
    raise ValueError(
      "Issue while invoking %r, result was:\n%s" % (args, result)
    )
  return popen.returncode, result


def ensureDir(path, mode):
  try:
    os.makedirs(path, mode)
  except OSError as e:
    if e.errno == errno.EEXIST and os.path.isdir(path):
      os.chmod(path, mode)
    else:
      raise


def addressFromNetwork(network, index):
  addr = network[index]
  return ipaddress.ip_interface((addr, network.prefixlen))


# Wrapping, Tracing & Dry-running

class WrappedModule(object):
  def __init__(self, conf, module, **methods):
    self._conf = conf
    self._module = module
    add = self._addWrapper
    for name, fake_result in methods.items():
      add(name, fake_result)

  def _addWrapper(self, name, fake_result):
    logger = self._conf.logger
    if self._conf.dry_run:
      def wrapper(*args, **kw):
        l = [repr(x) for x in args] + ['%s=%r' % item for item in kw.items()]
        logger.debug('%s(%s)' % (name, ', '.join(l)))
        return fake_result
    else:
      def wrapper(*args, **kw):
        l = [repr(x) for x in args] + ['%s=%r' % item for item in kw.items()]
        logger.debug('%s(%s)' % (name, ', '.join(l)))
        return getattr(self._module, name)(*args, **kw)
    setattr(self, name, wrapper)

  def __getattr__(self, name):
    return getattr(self._module, name)


class WrappedSystem(object):
  def __init__(self, conf):
    self.conf = conf
    self.call = call
    self.os = os
    self.pwd = pwd
    self.abort = conf.abort

  def __enter__(self):
    global call, os, pwd
    conf = self.conf
    os = WrappedModule(
      conf,
      os,
      chmod=None,
      chown=None,
      makedirs=None,
      mkdir=None,
    )
    pwd = WrappedModule(
      conf,
      pwd,
      getpwnam=type('fake_getpwnam', (), {'pw_uid': 12345, 'pw_gid': 54321}),
    )
    if conf.dry_run:
      conf.abort = conf.dryError
      def dry_call(args, throw=True):
        conf.logger.debug(' '.join(args))
        it = iter(args)
        if all(w in it for w in ('ip', 'addr', 'show')): # subsequence test
          return self.call(args, throw=False)
        return 0, ''
      call = dry_call
    else:
      def tracing_call(args, throw=True):
        conf.logger.debug(' '.join(args))
        return self.call(args, throw)
      call = tracing_call

  def __exit__(self, *exc):
    global call, os, pwd
    call = self.call
    os = self.os
    pwd = self.pwd
    self.conf.abort = self.abort
    return False
